/*
 * Copyright (c) 1991-94 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the name of Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
 * POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*------------------------------------------------------------
 *  This is an example from The Inventor Mentor,
 *  chapter 17, example 2.
 *
 *  Example of combining Inventor and OpenGL rendering.
 *  Create an Inventor render area and draw a red cube 
 *  and a blue sphere.  Render the floor with OpenGL 
 *  through a Callback node.
 *----------------------------------------------------------*/

#include <GL/gl.h>
#include <Inventor/SbLinear.h>
#include <Inventor/Xt/SoXt.h>
#include <Inventor/Xt/SoXtRenderArea.h>
#include <Inventor/Xt/viewers/SoXtExaminerViewer.h>
#include <Inventor/projectors/SbPlaneProjector.h>
#include <Inventor/nodes/SoCallback.h>
#include <Inventor/nodes/SoBaseColor.h>
#include <Inventor/events/SoMouseButtonEvent.h>
#include <Inventor/nodes/SoEventCallback.h>
#include <Inventor/nodes/SoCube.h>
#include <Inventor/nodes/SoCylinder.h>
#include <Inventor/nodes/SoDirectionalLight.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoPerspectiveCamera.h>
#include <Inventor/nodes/SoSphere.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoTransform.h>
#include <iostream.h>
#include "AliveVideo.h"
#include "Gesture.h"
#include "serial.hh"
#include "VideoMatte.h"

#define TIMELASTTHRESHOLD 20 
#define ARMTOHEADRATIO .4
/* duration for a bend, and % of nominal height needed to pass test  */
#define BENDD 5    
#define BENDT .75
#define ARMTOHEADRATIO .4
#define NW .33
#define OW .66

AliveVideo::AliveVideo(SoPerspectiveCamera *cam, SoXtRenderArea *ra, 
	SbVec3f camPos,SbMatrix camO,float vscale):
	AliveCamera(SoPerspectiveCamera *cam, SoXtRenderArea *ra, 
	SbVec3f camPos,SbMatrix camO,float vscale)
{
    int i;
    connection = -1;
    showBodyMarkers = 0;
    showVideoMatte = 0;
    for(i=0;i<4;i++){
	markerXF.append(new SoTransform);
    }
    /* open the connection */
    connection = -1;
    openConnection();
    if(!connection)
        printf("WARNING DID NOT CONNECT \n");
    timeCounter = 0;
}

void AliveVideo::setShowBodyMarkers(int yn, SoSeparator *where)
{
    if(yn){
	buildPersonTracker(where);
    }
    showBodyMarkers = yn;
}

void AliveVideo::setShowVideoMatte(int yn, SoSeparator *where)
{
    if(yn){
	videoMatte = new VideoMatte(this, where);
    }
    showVideoMatte = yn;
}

void AliveVideo::buildPersonTracker(SoSeparator *theMarkers)
{
    int i;
    SoSeparator *sep;
    SoSeparator *rsep = new SoSeparator;
    
    SoBaseColor *red = new SoBaseColor;
    red->rgb.setValue(1, 0, 0);
    rsep->addChild(red);
    SoSphere *sp = new SoSphere;
    sp->radius.setValue(.5);
    
    for(i=0;i<4;i++){
	sep = new SoSeparator;
	rsep->addChild(sep);
	sep->addChild((SoTransform *)markerXF[i]);
	sep->addChild(sp);
    }
    theMarkers->addChild(rsep);
}

void AliveVideo::projectPersonInto3DWorld()
{
    SbVec3f wp;
    SbVec3f pt3d;
    SbVec2f pt2d;
    /* first find z */
    SbVec2f projOnFloor((float)(bbox[2]-bbox[0])/(2.0*COGW),
	(float)bbox[3]/COGH);
    projectYtoZFloor(projOnFloor,wp);
    
    /* this adjustment to compensate for length of feet */
    wp[2]-= 6.0/vXwScale;
    //printf("feet at z => %f \n", wp[2]);
    
    /* now do projections for body parts */
    pt2d.setValue((float)rhxy[0]/COGW,(float)rhxy[1]/COGH);
    projectScreenYtoWorldY(pt2d,rhand,wp[2]);
    
    pt2d.setValue((float)lhxy[0]/COGW,(float)lhxy[1]/COGH);
    projectScreenYtoWorldY(pt2d,lhand,wp[2]);
    
    pt2d.setValue((float)hxy[0]/COGW,(float)hxy[1]/COGH);
    projectScreenYtoWorldY(pt2d,head,wp[2]);
    
    pt2d.setValue((float)cxy[0]/COGW,(float)cxy[1]/COGH);
    projectScreenYtoWorldY(pt2d,cm,wp[2]);
    
    /* do projection of bounding box (bottom left and top right */
    pt2d.setValue((float)bbox[0]/COGW,(float)bbox[1]/COGH);
    projectScreenYtoWorldY(pt2d,bbox3d[0],wp[2]);
    
    pt2d.setValue((float)bbox[2]/COGW,(float)bbox[3]/COGH);
    projectScreenYtoWorldY(pt2d,bbox3d[1],wp[2]);
    
    if(showBodyMarkers){
	((SoTransform *)markerXF[2])->translation.setValue(head);
	((SoTransform *)markerXF[1])->translation.setValue(lhand);
	((SoTransform *)markerXF[0])->translation.setValue(rhand);
	((SoTransform *)markerXF[3])->translation.setValue(cm);
    }
    if(showVideoMatte)
	videoMatte->updateMatte_sgi();
	
    /* now check if this is a new person, if so, get nominal height 
    if(timeCounter-timeLastSeen > TIMELASTTHRESHOLD)
	newPerson();
    timeLastSeen = timeCounter;
    testGestures();
    */
}

void AliveVideo::printDataFromCognex()
{
    cout << "bounding box: "
	<< bbox[0] << " "<< bbox[1] << " "<< bbox[2] << " "<< bbox[3] <<"\n";
    cout << "center of mass: "
	<< cxy[0] <<  " " <<cxy[1]<<"\n";
    cout << "right hand: "
	<< rhxy[0] << " " << rhxy[1]<<"\n";
    cout << "left hand: "
	<< lhxy[0] <<  " " <<lhxy[1]<<"\n";
    cout << "head: "
	<< hxy[0] <<  " " << hxy[1]<<"\n";
}

int AliveVideo::getNewPosition(SbVec3f *cm1, SbVec3f *lh,SbVec3f *rh, SbVec3f *hd)
{
    if(readDataFromCognex()){
	*cm1 = cm;
	*lh = lhand;
	*rh = rhand;
	*hd = head;
	return 1;
    }
    else
	return 0;
}

void AliveVideo::smooth()
{
    int i;
    for(i=0;i<2;i++){
	oldcxy[i] = cxy[i] = (NW*cxy[i] + OW*oldcxy[i]);
	oldrhxy[i] =rhxy[i] = (NW*rhxy[i] + OW*oldrhxy[i]);
	oldlhxy[i] =lhxy[i] = (NW*lhxy[i] + OW*oldlhxy[i]);
	oldhxy[i] =hxy[i] = (NW*hxy[i] + OW*oldhxy[i]);
    }
}

int  AliveVideo::readDataFromCognex()
{
    int isOne;
     
     timeCounter++;
     if(!connection){
	 printf("no connection");
	 openConnection();
	 return 0;
     }
     isOne = get_packet(connection, &response);
     if(isOne){
        
	//checkForRestart();
	bbox[0] = (int)response.bb.xmin;
        bbox[2] = (int)response.bb.xmax;
        bbox[1] = (int)COGH - (int)response.bb.ymin;
        bbox[3] = (int)COGH - (int)response.bb.ymax;
	
	cxy[0] = (int)response.cm.x;
	cxy[1] = (int)COGH - (int)response.cm.y;
	rhxy[0] = (int)response.hand0.x;
	rhxy[1] = (int)COGH - (int)response.hand0.y;
	lhxy[0] = (int)response.hand1.x;
	lhxy[1] = (int)COGH - (int)response.hand1.y;
	hxy[0] = (int)response.head.x;
	hxy[1] = (int)COGH - (int)response.head.y;
	
	//smooth();
	//printDataFromCognex();
	projectPersonInto3DWorld();
	return 1;
    }
    else{
	//printf("no data from cognex \n");
	return 0;
    }
}

void AliveVideo::openConnection()
{
    if(connection<0){
	connection = connect_serial();
	firstTime = 1;
    }
}

void AliveVideo::closeConnection()
{
    if(connection>=0)
	close_serial(connection);
    printf("connection closed\n");
    connection = -1;
}


