/*
 * Copyright (c) 1991-94 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the name of Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
 * POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*------------------------------------------------------------
 *  This is an example from The Inventor Mentor,
 *  chapter 17, example 2.
 *
 *  Example of combining Inventor and OpenGL rendering.
 *  Create an Inventor render area and draw a red cube 
 *  and a blue sphere.  Render the floor with OpenGL 
 *  through a Callback node.
 *----------------------------------------------------------*/

#include <GL/gl.h>
#include <Inventor/SbLinear.h>
#include <Inventor/Xt/SoXt.h>
#include <Inventor/Xt/SoXtRenderArea.h>
#include <Inventor/Xt/viewers/SoXtExaminerViewer.h>
#include <Inventor/projectors/SbPlaneProjector.h>
#include <Inventor/nodes/SoCallback.h>
#include <Inventor/nodes/SoBaseColor.h>
#include <Inventor/events/SoMouseButtonEvent.h>
#include <Inventor/nodes/SoEventCallback.h>
#include <Inventor/nodes/SoCube.h>
#include <Inventor/nodes/SoCylinder.h>
#include <Inventor/nodes/SoDirectionalLight.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoPerspectiveCamera.h>
#include <Inventor/nodes/SoOrthographicCamera.h>
#include <Inventor/nodes/SoSphere.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/actions/SoGetMatrixAction.h>
#include <iostream.h>
#include <fstream.h>
#include "AliveCamera.h"
#include "TimeStat.h"
AliveCamera::AliveCamera(SoCamera *cam, SoXtRenderArea *ra)
{
    commonInit(cam, ra);
}

AliveCamera::AliveCamera(SoCamera *cam, SoXtRenderArea *ra, 
	SbVec3f camPos,SbMatrix camO,float vscale)
{
    commonInit(cam, ra);
    vXwScale = vscale;
    adjustCamera(camPos, camO, vscale, SbVec2s(640, 480));
}

void AliveCamera::commonInit(SoCamera *cam, SoXtRenderArea *ra)
{
    camera = cam;
    markers = NULL;
    renderArea = ra;
    ((SoXtExaminerViewer *)renderArea)->setAutoClipping(0);
    showMouseMarkers = 0;
}

void AliveCamera::setShowMouseMarkers(int yn, SoSeparator *where)
{
    if(yn){
	markers = where;
    }
    showMouseMarkers = yn;
}

void AliveCamera::addMarker(SbVec2f xy)
{
    SbViewVolume vv;
    vv = camera->getViewVolume();
    SbVec3f mp = vv.getPlanePoint(camera->nearDistance.getValue() + 1.0, xy);
    SoSeparator *sep = new SoSeparator;
    SoTransform *xf = new SoTransform;
    xf->translation.setValue(mp);
    sep->addChild(xf);
    SoBaseColor *red = new SoBaseColor;
    red->rgb.setValue(1.0, 0, 0);
    sep->addChild(red);
    SoSphere *sp = new SoSphere;
    sp->radius.setValue(.06);
    sep->addChild(sp);
    markers->addChild(sep);
}

void AliveCamera::printCameraParameters()
{
    SbVec3f pos = camera->position.getValue();
    cout << "position\t";
    cout << pos[0]<< "\t" << pos[1] <<"\t" <<  pos[2] << "\n";
    
    SbMatrix mat, aff,  proj;
    SbRotation rot = camera->orientation.getValue();
    rot.getValue(mat);
    int i;
    cout << "orientation\n";
    for(i=0;i<3;i++)
	cout << mat[i][0] <<"\t" << mat[i][1]<< "\t" << mat[i][2] <<"\n";
    cout << "focal distance \t" 
	<<(camera)->focalDistance.getValue() <<"\n";
    if(camera->getTypeId() == SoPerspectiveCamera::getClassTypeId())
	cout << "height angle \t" 
	    <<((SoPerspectiveCamera *)camera)->heightAngle.getValue() <<"\n";
    else
	cout << "height \t" 
	    <<((SoOrthographicCamera *)camera)->height.getValue() <<"\n";
    cout << "aspect ratio \t" 
	<<(camera)->aspectRatio.getValue() <<"\n";
    
    SbVec2s we, wo, ws;
    wo.setValue(0, 0);
    we = renderArea->getSize();
    cout << "renderArea size \t" 
    <<"\t"<<we[0]
    << "\t"<<we[1]
    <<"\n";
    
    SbViewportRegion vpr(we[0], we[1]);
    vpr=((SoPerspectiveCamera *)camera)->getViewportBounds(vpr);
    wo = vpr.getViewportOriginPixels();
    we = vpr.getViewportSizePixels();
    ws = vpr.getWindowSize();
    cout << "viewport mapping \t" 
    <<wo[0] 
    << "\t"<<wo[1]
    << "\t"<<we[0]
    << "\t"<<we[1]
    << "\t"<<ws[0]
    << "\t"<<ws[1]
    <<"\n";
	
    // Now do view volume stuff 
    SbViewVolume vv;
    vv = camera->getViewVolume();
    cout << "view volume parameters \n"
	<< " near->" << vv.getNearDist() <<"\n"
	<< " width->" << vv.getWidth() <<"\n"
	<< " height->" << vv.getHeight() <<"\n"
	<< " depth->" << vv.getDepth() <<"\n";

    vv.getMatrices(aff, proj);
    cout << "affine\n";
    for(i=0;i<4;i++)
	cout << aff[i][0] <<"\t" << aff[i][1]<< "\t" << aff[i][2] 
	<< "\t" << aff[i][3]<<"\n";
    cout << "projective\n";
    for(i=0;i<4;i++)
	cout << proj[i][0] <<"\t" << proj[i][1]<< "\t" << proj[i][2] 
	<< "\t" << proj[i][3]<<"\n";
    aff = vv.getMatrix();
    cout << "viewing matrix\n";
    for(i=0;i<4;i++)
	cout << aff[i][0] <<"\t" << aff[i][1]<< "\t" << aff[i][2] 
	<< "\t" << aff[i][3]<<"\n";

    cout.flush();

}


/*
 * The following function projects a 2D screen point to a 3D point on
 * the floor of the virtual world in the scale of the virtual world.
 * Multiply by the world-to-virtual scale factor to get the actual 
 * world point 
 */
void AliveCamera::projectYtoZFloor(SbVec2f pt2d,SbVec3f &pt3d)
{
    SbPlaneProjector pp(SbPlane(SbVec3f(0, 1, 0), 0), 0);
    SbViewVolume vv;
    
    pp.setViewVolume(camera->getViewVolume());
    pt3d = pp.project(pt2d);
/*
    cout << "projected point in real world\n"
	<< " x->" << pt3d[0] <<"\n"
	<< " y->" << pt3d[1] <<"\n"
	<< " z->" << pt3d[2] <<"\n";
    cout.flush();
*/    
}

/*
 * The following function projects a 2D screen point to a 3D point on
 * a vertical plane positioned at Z in the virtual world 
 * in the scale of the virtual world.Multiply by the world-to-virtual 
 * scale factor to get the actual world point 
 */
void AliveCamera::projectYtoZFloor(float *pt2df,float *pt3df)
{
    SbVec2f pt2d(pt2df[0],1.0-pt2df[1]);
    SbVec3f pt3d;
    SbPlaneProjector pp(SbPlane(SbVec3f(0, 1, 0), 0), 0);
    SbViewVolume vv;
    
    pp.setViewVolume(camera->getViewVolume());
    pt3d = pp.project(pt2d);
    pt3d.getValue(pt3df[0], pt3df[1], pt3df[2]);
}

void AliveCamera::projectScreenYtoWorldY(SbVec2f pt2d, SbVec3f &pt3d,float Z)
{
  //pt2d[1] = 1.0 - pt2d[1];
  SbPlaneProjector pp(SbPlane(SbVec3f(0, 0, 1),Z), 0);
  SbViewVolume vv;
  
  pp.setViewVolume(camera->getViewVolume());
  pt3d = pp.project(pt2d);
}

void AliveCamera::projectScreenYtoWorldY(float *pts,int npts,int stride,  
    SbVec3f *pt3d,float Z)
{
    int i;
    
    SbPlaneProjector pp(SbPlane(SbVec3f(0, 0, 1),Z), 0);
    SbViewVolume vv;
    
    pp.setViewVolume(camera->getViewVolume());
    for(i=0;i<npts*stride;i+=stride){
	pt3d[i/stride] = pp.project(SbVec2f(pts[i],1- pts[i+1]));
    }
}

void AliveCamera::projectScreenYtoWorldY(float *pts,int npts,int stride,  
    float *pt3ds,float Z)
{
    int i;
    SbVec3f pt3d;
    
    SbPlaneProjector pp(SbPlane(SbVec3f(0, 0, 1),Z), 0);
    SbViewVolume vv;
    
    pp.setViewVolume(camera->getViewVolume());
    for(i=0;i<npts*stride;i+=stride){
	int j = i/stride;
	pt3d = pp.project(SbVec2f(pts[i],1- pts[i+1]));
	pt3d.getValue(pt3ds[3*j],pt3ds[3*j+1], pt3ds[3*j+2]); 
    }
}

void AliveCamera::trackMouseEvents()
{
    SoEventCallback *myEventCB = new SoEventCallback;
        myEventCB->addEventCallback(
         SoMouseButtonEvent::getClassTypeId(),
         AliveCamera::AliveCameraCB,
         this);
    if(markers)
	markers->addChild(myEventCB);
    else{
	printf("Need a marker");
    } 
}

class BackProject {
    public:
    SbMatrix spm_static;
    SbMatrix spm1;
    SoTransform *xf;
    SbVec3f cp;
    SbMatrix matrix;
    SbMatrix m2;
    SbPlane plane;
    float planeZ;
    SbMatrix spm;
    SbMatrix proj;
    BackProject(SoPerspectiveCamera *cam);
    void setProjectionPlane(float z);
    void backProjectPt(const SbVec3f &sp1, SbVec3f &projPt, float z);
};

BackProject::BackProject(SoPerspectiveCamera *cam)
{
    xf = new SoTransform();
    xf->rotation.setValue(cam->orientation.getValue());
    cp = cam->position.getValue();
    xf->translation.setValue(cp[0],cp[1],0);
    SoGetMatrixAction gma(SbViewportRegion());
    gma.apply(xf);
    matrix = gma.getMatrix();
    m2 = gma.getInverse();
    
    // sm is a matrix which shifts the screen point (accounting for
    // 0 to 1 viewport mapping, and positions it in z. Currently,
    // the z is a fudge factor that depends on the camera parameters
    
    spm_static.makeIdentity();
    spm_static.setTranslate(SbVec3f(-.5, -.5, -1.2*.78/(cam->heightAngle.getValue())));
    spm1.makeIdentity();
    spm1.setScale(SbVec3f(4.0/3.0, 1, 1));
    spm_static.multRight(spm1);
    planeZ = 0;
}

void BackProject::setProjectionPlane(float z)
{
    planeZ = z;
    plane = SbPlane(SbVec3f(0, 0, 1), -z+cp[2]);
    plane.transform(m2);
    SbVec3f norm = plane.getNormal();
    float d = plane.getDistanceFromOrigin();
    
    //Now build the projection matrix
    proj = SbMatrix(
	-d, 0,  0,  norm[0],
	0, -d,  0,  norm[1],
	0, 0,  -d,  norm[2],
	0, 0,  0, 0); 
    
    // Now concatenate up the matrix
    // sm * proj * matrix
    // The first guy converts the point to the camera coordinates from
    // the screen coordinates. The second guy projects it onto the projection
    // plane which has been transformed into the camera's frame. The third
    // guy takes the result and converts it back into the world coordinates
    // spm can be calculated once per projection plane
    spm = spm_static;
    spm.multRight(proj);
    spm.multRight(matrix);
    
    SbMatrix spfinal;
    spfinal.identity();
    spfinal.setTranslate(SbVec3f(0, 0, cp[2]));
    spm.multRight(spfinal);
}

void BackProject::backProjectPt(const SbVec3f &sp1, SbVec3f &projPt, float z)
{
    if(fabs(z-planeZ)> .1)
	setProjectionPlane(z);
	
    SbVec3f sp = sp1;
    sp[2] = 0;
    // This does the actual conversion
    spm.multVecMatrix(sp, projPt);
    //printf("%f %f %f \n", projPt[0], projPt[1], projPt[2]);
    return;
}

void convertPt(const SbVec3f &sp1, SbVec3f &projPt, SoPerspectiveCamera *cam, 
    float z)
{    
    // This is specifying the coordinate system in which you will
    // do the actual shadow calculations. This need only be done
    // once.
    SbVec3f sp = sp1;
    static SbMatrix spm_static;
    static SbMatrix spm1;
    static SoTransform *xf;
    static SbVec3f cp;
    static SbMatrix matrix;
    static SbMatrix m2; 
    static int ft = 1;
    if(ft==1){
	xf = new SoTransform();
	xf->rotation.setValue(cam->orientation.getValue());
	cp = cam->position.getValue();
	xf->translation.setValue(cp[0],cp[1],0);
	SoGetMatrixAction gma(SbViewportRegion());
	gma.apply(xf);
	matrix = gma.getMatrix();
	m2 = gma.getInverse();
	
	// sm is a matrix which shifts the screen point (accounting for
	// 0 to 1 viewport mapping, and positions it in z. Currently,
	// the z is a fudge factor that depends on the camera parameters
	
	spm_static.makeIdentity();
	spm_static.setTranslate(SbVec3f(-.5, -.5, -1.2*.78/(cam->heightAngle.getValue())));
	spm1.makeIdentity();
	spm1.setScale(SbVec3f(4.0/3.0, 1, 1));
	spm_static.multRight(spm1);
	
	ft = 0;
    }
    
    // This builds the projection plane and transforms it so that you can
    // build the projection matrix
    SbPlane plane(SbVec3f(0, 0, 1), -z+cp[2]);
    plane.transform(m2);
    SbVec3f norm = plane.getNormal();
    float d = plane.getDistanceFromOrigin();
    
    //Now build the projection matrix
    SbMatrix proj(
	-d, 0,  0,  norm[0],
	0, -d,  0,  norm[1],
	0, 0,  -d,  norm[2],
	0, 0,  0, 0); 
    
    // make sure the z is 0
    sp[2] = 0;
    
    
    // Now concatenate up the matrix
    // sm * proj * matrix
    // The first guy converts the point to the camera coordinates from
    // the screen coordinates. The second guy projects it onto the projection
    // plane which has been transformed into the camera's frame. The third
    // guy takes the result and converts it back into the world coordinates
    // spm can be calculated once per projection plane
    SbMatrix spm = spm_static;
    spm.multRight(proj);
    spm.multRight(matrix);
    
    SbMatrix spfinal;
    spfinal.identity();
    spfinal.setTranslate(SbVec3f(0, 0, cp[2]));
    spm.multRight(spfinal);

    // This does the actual conversion
    spm.multVecMatrix(sp, projPt);
    printf("%f %f %f \n", projPt[0], projPt[1], projPt[2]);
    return;
}

// This routine is called for every mouse button event.
void AliveCamera::AliveCameraCB(void *userData, SoEventCallback *eventCB)
{
  AliveCamera *av = (AliveCamera *)userData;
  SoCamera *cam = av->camera;
  const SoMouseButtonEvent *event = (SoMouseButtonEvent *)eventCB->getEvent();
  static float lastZ;
  
  // Check for mouse button being pressed
  if (SO_MOUSE_PRESS_EVENT(event, BUTTON1)) {
    const SbViewportRegion &myRegion = 
      eventCB->getAction()->getViewportRegion();
    SbVec2s mp = event->getPosition(myRegion);
    
    eventCB->setHandled();
    //printCameraParameters(cam);
    SbVec2f nmp = event->getNormalizedPosition(myRegion);
    cout << "mouse position  \n"
	 << " x->" << nmp[0] <<"\n"
	 << " y->" << nmp[1] <<"\n";
    cout.flush();
    
    //addMarker(cam, nmp);
    SbVec3f p;
    av->projectYtoZFloor(nmp, p);
    lastZ = p[2];
    if(av->showMouseMarkers){
      SoSeparator *sep = new SoSeparator;
      SoTransform *xf = new SoTransform;
      xf->translation.setValue(p);
      sep->addChild(xf);
      SoBaseColor *red = new SoBaseColor;
      red->rgb.setValue(1, 0, 0);
      sep->addChild(red);
      SoSphere *sp = new SoSphere;
      sp->radius.setValue(.5);
      sep->addChild(sp);
      av->markers->addChild(sep);
      
      cout << "projected point in virtual world\n"
	   << " x->" << p[0] <<"\n"
	   << " y->" << p[1]  <<"\n"
	   << " z->" << p[2]  <<"\n";
      cout.flush(); 
    } 
  }
  else if (SO_MOUSE_PRESS_EVENT(event, BUTTON2)) {
    const SbViewportRegion &myRegion = 
      eventCB->getAction()->getViewportRegion();
    SbVec2s mp = event->getPosition(myRegion);
    eventCB->setHandled();
//    av->printCameraParameters();
    SbVec2f nmp = event->getNormalizedPosition(myRegion);
    cout << "mouse position \n"
	 << " x->" << nmp[0] <<"\n"
	 << " y->" << nmp[1] <<"\n";
    cout.flush();
    SbVec3f p;
    Timestat *ts1 = new Timestat("regular", 1000);
    Timestat *ts2 = new Timestat("new and improved", 1000);
    int tt;
    for(tt=0;tt<1000;tt++){
	ts1->startTime();
	av->projectScreenYtoWorldY(nmp, p, lastZ);
	ts1->endTime();
    }
    SbVec3f tmpt(nmp[0],nmp[1], 0);
    SbVec3f newpt;
   // convertPt(tmpt, newpt, (SoPerspectiveCamera *)av->camera, lastZ);
    static BackProject *bp = NULL;
    if(!bp)
	bp = new BackProject((SoPerspectiveCamera *)av->camera);
   // bp->setProjectionPlane(lastZ);
    for(tt=0;tt<1000;tt++){
	ts2->startTime();
	bp->backProjectPt(tmpt, newpt, lastZ);
	ts2->endTime();
    }
    printf("%f \n", (float)(ts1->avgTime/ts2->avgTime));
    printf("%f %f \n", newpt[0]/p[0], newpt[1]/p[1]); 
    if(av->showMouseMarkers){
      SoSeparator *sep = new SoSeparator;
      SoTransform *xf = new SoTransform;
      xf->translation.setValue(newpt);
      sep->addChild(xf);
      SoBaseColor *green = new SoBaseColor;
      green->rgb.setValue(0, 1, 0);
      sep->addChild(green);
      SoSphere *sp = new SoSphere;
      sp->radius.setValue(.5);
      sep->addChild(sp);
      av->markers->addChild(sep);
      
      cout << "projected point in virtual world\n"
	   << " x->" << p[0] <<"\n"
	   << " y->" << p[1]  <<"\n"
	   << " z->" << p[2]  <<"\n";
      cout.flush();
    }
/*
    {
      static num_points =0;
      static matrix_type *test = make_matrix_type(5,30,"test");
      float x,y,z;
      int ret = scanf("Global position (x,y,z): %f %f %f",&x,&y,&z);
      if (ret == 3) {
	MREF(test, num_points,0) = x;
	MREF(test, num_points,1) = y;
	MREF(test, num_points,2) = z;
	MREF(test, num_points,3) = nmp[0];
	MREF(test, num_points,4) = nmp[1];
	if (num_points > 3) {
	  matrix_type *camera;
	  do_calibration(num_points,test,&camera);
	  
	  free_matrix_type(camera);
	}
      }
    }
*/
  } 
}

XVisualInfo *AliveCamera::buildEnabledVisual(int *attributeList)
{
    XVisualInfo	*vi;
    Display *display;
    
    // Open the X display 
    display = XOpenDisplay(0);
    
    // Initialize the GLX visual and context
    vi = glXChooseVisual(display, DefaultScreen(display), attributeList);
    if(!vi)
	printf("trouble in X land \n");
    return vi;
}

void AliveCamera::readCameraParametersFromFile(char *fn)
{
    int i, j;
    float dp[3], ha, sc;
    float sx, sy;
    float mat[3][3];
    ifstream fin(fn);
    if(!fin)
	printf("camera description not found \n");
    else{
	fin >> sx;
	fin >> sy;
	// position of camera
	for(j=0;j<3;j++){
	    fin >> dp[j];
	}
	// rotation matrix
	for(i=0;i<3;i++)
	    for(j=0;j<3;j++)
		fin>>mat[i][j];
	
	// height angle
	fin>>ha;
	
	// scale
	fin>> sc; 
	
	SbMatrix rm(mat[0][0],mat[0][1], mat[0][2], 0, 
	     mat[1][0],mat[1][1], mat[1][2], 0,
	     mat[2][0],mat[2][1], mat[2][2], 0,
	     0,0, 0, 1);
	SbVec3f disp(dp[0],dp[1], dp[2]); 
	adjustCamera(disp, rm, sc, SbVec2s(sx, sy));
	if(camera->getTypeId() == SoPerspectiveCamera::getClassTypeId())
	    ((SoPerspectiveCamera *)camera)->heightAngle  = ha;
	else
	    ((SoOrthographicCamera *)camera)->height  = ha;
	printCameraParameters();
    }	
}
 
void AliveCamera::adjustCamera(SbVec3f disp,SbMatrix rm, float vscale, SbVec2s size)
{
    //SbVec2s size = SbVec2s(640, (short)(.75*640));
    vXwScale = vscale;
    disp /= vXwScale;
    camera->aspectRatio.setValue( SO_ASPECT_VIDEO );
    camera->viewportMapping.setValue(SoCamera::CROP_VIEWPORT_FILL_FRAME);
    camera->position.setValue(disp);
    camera->orientation.setValue(SbRotation(rm));
    if(camera->getTypeId() == SoPerspectiveCamera::getClassTypeId())
	    ((SoPerspectiveCamera *)camera)->heightAngle  = .704;
	else
	    ((SoOrthographicCamera *)camera)->height  = .704;  
    camera->nearDistance = 1.0;
    camera->farDistance  = 220.0;
   
    renderArea->setSize(size);
}

void AliveCamera::convertToLUV(SbColor color)
{
    float y, u, v;
    
    y = .3*color[0] + .59*color[1] + .11*color[2];
    u = .493*(color[1]-y);
    v = .897*(color[0]-y);
    printf("%f %f %f\n", y, u, v);
    printf("%f %f %f\n", y*255, u*255, v*255);
}
