#include <matrix.h>
#include <stdio.h>

#define MLHS(mat, y, x) (*((mat)->m + (y) * (mat)->w + (x)))

#undef VERBOSE


matrix_type *buildMatrix(n,test)
     int n;
     matrix_type *test;
{
  matrix_type *A = make_matrix_type(12,2*n,"A");
  int i;
  for (i=0;i<n;i++) 
    {
      FLOATTYPE m1 = MREF(test,i,0);
      FLOATTYPE m2 = MREF(test,i,1);
      FLOATTYPE m3 = MREF(test,i,2);
      FLOATTYPE u  = MREF(test,i,3);
      FLOATTYPE v  = MREF(test,i,4);
      MREF(A,i*2+0, 0) = m1;
      MREF(A,i*2+0, 1) = m2;
      MREF(A,i*2+0, 2) = m3;
      MREF(A,i*2+0, 3) = 1.0;
      MREF(A,i*2+0, 4) = 0.0;
      MREF(A,i*2+0, 5) = 0.0;
      MREF(A,i*2+0, 6) = 0.0;
      MREF(A,i*2+0, 7) = 0.0;
      MREF(A,i*2+0, 8) = -u*m1;
      MREF(A,i*2+0, 9) = -u*m2;
      MREF(A,i*2+0,10) = -u*m3;
      MREF(A,i*2+0,11) = -u;
      MREF(A,i*2+1, 0) = 0.0;
      MREF(A,i*2+1, 1) = 0.0;
      MREF(A,i*2+1, 2) = 0.0;
      MREF(A,i*2+1, 3) = 0.0;
      MREF(A,i*2+1, 4) = m1;
      MREF(A,i*2+1, 5) = m2;
      MREF(A,i*2+1, 6) = m3;
      MREF(A,i*2+1, 7) = 1.0;
      MREF(A,i*2+1, 8) = -v*m1;
      MREF(A,i*2+1, 9) = -v*m2;
      MREF(A,i*2+1,10) = -v*m3;
      MREF(A,i*2+1,11) = -v; 
    }
#ifdef VERBOSE
  matrix_print(stdout,A);
#endif
  return (A);
}

matrix_type *extractC(A)
     matrix_type *A;
{
  matrix_type *C = make_matrix_type(9,matrix_height(A),"C");
  int i;
  for (i=0;i<matrix_height(A);i++) 
    {
      MREF(C,i, 0) = MREF(A,i,0);
      MREF(C,i, 1) = MREF(A,i,1);
      MREF(C,i, 2) = MREF(A,i,2);
      MREF(C,i, 3) = MREF(A,i,3);
      MREF(C,i, 4) = MREF(A,i,4);
      MREF(C,i, 5) = MREF(A,i,5);
      MREF(C,i, 6) = MREF(A,i,6);
      MREF(C,i, 7) = MREF(A,i,7);
      MREF(C,i, 8) = MREF(A,i,11);
    }
#ifdef VERBOSE
  matrix_print(stdout,C);
#endif
  return (C);
}

matrix_type *extractD(A)
     matrix_type *A;
{
  matrix_type *D = make_matrix_type(3,matrix_height(A),"D");
  int i;                                                                                                                                           
  for (i=0;i<matrix_height(A);i++) 
    {
      MREF(D,i, 0) = MREF(A,i,8);
      MREF(D,i, 1) = MREF(A,i,9);
      MREF(D,i, 2) = MREF(A,i,10);
    }
#ifdef VERBOSE
  matrix_print(stdout,D);
#endif
  return (D);
}

matrix_type *make_ctci(C)
     matrix_type *C;
{
  /* ctci = (c' * c)^(-1); */
  matrix_type *ctc= make_matrix_type(matrix_width(C),
				     matrix_width(C),
				     "CTC");
  matrix_type *ctci= make_matrix_type(matrix_width(C),
				      matrix_width(C),
				      "CTCI");
  
  matrix_transpose_multiply(C,C,ctc);
  
  matrix_invert(ctc,ctci);
  
  free_matrix_type( ctc );
  
#ifdef VERBOSE
  matrix_print(stdout,ctci);
#endif
  return (ctci);
}

matrix_type *make_eigsource(C,D,CTCI)
     matrix_type *C;
     matrix_type *D;
     matrix_type *CTCI;
{
  /* ea = (d'*d - d'*c*(ctci*c'*d) */
  matrix_type *dtd = make_matrix_type(matrix_width(D),
				      matrix_width(D),
				      "DTD");
  matrix_type *dtc = make_matrix_type(matrix_width(C),
				      matrix_width(D),
				      "DTC");
  matrix_type *ctd = make_matrix_type(matrix_width(D),
				      matrix_width(C),
				      "CTD");
  
  matrix_transpose_multiply(D,D,dtd);
  matrix_transpose_multiply(C,D,ctd);
  matrix_transpose_multiply(D,C,dtc);
  {
    matrix_type *ctcictd = make_matrix_type(matrix_width(ctd),
					    matrix_height(CTCI),
					    "CTCICTD");
    matrix_multiply(CTCI,ctd,ctcictd);
    
    {
      matrix_type *dtcctcictd = make_matrix_type(matrix_width(ctcictd),
						 matrix_height(dtc),
						 "DTCCTCICTD");
      matrix_multiply(dtc,ctcictd,dtcctcictd);
      
      {
	
	matrix_type *eigsource = make_matrix_type(matrix_height(dtd),
						  matrix_width(dtd),
						  "EIGSOURCE");
	matrix_subtract(dtd,dtcctcictd,eigsource);
	
	free_matrix_type( dtd );
	free_matrix_type( ctd );
	free_matrix_type( dtc );
	free_matrix_type( ctcictd );
	free_matrix_type( dtcctcictd );
	
#ifdef VERBOSE
	matrix_print(stdout,eigsource);
#endif
	return (eigsource);
	
      }
    }
  }
}

matrix_type *find_z(ea)
     matrix_type *ea;
{
  matrix_type *v = make_matrix_type(matrix_height(ea),
				    matrix_width(ea),
				    "EIGVECTORS");
  matrix_type *d = make_matrix_type(matrix_width(ea),
				    1,
				    "EIGVALUES");
  {
    FLOATTYPE *D; 
    FLOATTYPE **V;
    FLOATTYPE **EA;
    int nrot;
    int i,j;
    
    
    EA = (FLOATTYPE **) malloc(sizeof(FLOATTYPE *) * (matrix_height(ea)+1));
    for(i=0;i<matrix_height(ea);i++)  
      {
	EA[i+1] = (FLOATTYPE *) malloc(sizeof(FLOATTYPE) * 
				       (matrix_width(ea)+1));
	for(j=0;j<matrix_width(ea);j++) 
	  EA[i+1][j+1] = MREF(ea,i,j);
      }
    
    V = (FLOATTYPE **) malloc(sizeof(FLOATTYPE *) * (matrix_height(v)+1));
    for(i=0;i<matrix_height(v);i++)  
      {
	V[i+1] = (FLOATTYPE *) malloc(sizeof(FLOATTYPE) * 
				      (matrix_width(v)+1));
      }
    
    D = (FLOATTYPE *) malloc(sizeof(FLOATTYPE) * (matrix_width(v) + 1));
    
    /* [x,ev] =eig(ea) */
    jacobi(EA,3,D,V,&nrot);
    
    /* evs = sort(ev) */
    eigsrt(D,V,3);
    
    for(i=0;i<matrix_height(v);i++) 
      for(j=0;j<matrix_width(v);j++) 
	MREF(v,i,j) = V[i+1][j+1];
    for(i=0;i<matrix_width(d);i++)  
      MREF(d,0,i) = D[i+1];
    
    for(i=0;i<matrix_height(ea);i++) free(EA[i+1]);
    free(EA);    
    for(i=0;i<matrix_height(v);i++) free(V[i+1]);
    free(V);    
    free(D);    
  }
  
#ifdef VERBOSE
  matrix_print(stdout,d);
  matrix_print(stdout,v);
#endif
  
  {
    matrix_type *z = make_matrix_type(1, 
				      matrix_height(v),
				      "Z");
    int i;

    /* [y,j] = min(evs(3,:)) */
    FLOATTYPE y = MREF(d,0,matrix_width(d)-1);
    
    /* z = x(:,j) */
    for (i=0;i<matrix_height(v);i++)
      MREF(z,i,0) = MREF(v,i,matrix_width(v)-1);
    
#ifdef VERBOSE
    matrix_print(stdout,z);
    printf ("(Y = %f)\n",y);
#endif
    
    free_matrix_type(d);
    free_matrix_type(v);

    return(z);
  }
}

matrix_type *find_y(ctci,c,d,z)
     matrix_type *ctci;
     matrix_type *c;
     matrix_type *d;
     matrix_type *z;
{
  /* % now find y */
  /* y = - ctci*c'*d*z; */
  
  matrix_type *ctcict= make_matrix_type(matrix_height(c),
					matrix_height(ctci),
					"CTCICT");
  matrix_multiply_transpose(ctci,c,ctcict);
  
#ifdef VERBOSE
      matrix_print(stdout,ctci);
      matrix_print(stdout,c);
      matrix_print(stdout,d);
      matrix_print(stdout,z);
#endif
  {
    matrix_type *ctcictd= make_matrix_type(matrix_width(d),
					   matrix_height(ctcict),
					   "CTCICTD");
    matrix_multiply(ctcict,d,ctcictd);
    
    {
      matrix_type *y= make_matrix_type(matrix_width(z),
				       matrix_height(ctcictd),
				       "Y");

      matrix_multiply(ctcictd,z,y);
      matrix_scale(y,-1.0,y);
      
#ifdef VERBOSE
      matrix_print(stdout,y);
#endif

      free_matrix_type( ctcict );
      free_matrix_type( ctcictd );
    
      return (y);
    }
  }
}

matrix_type *find_camera(y,z)
     matrix_type *y;
     matrix_type *z;
{     
  /* % P is the camera matrix */

  matrix_type *P = make_matrix_type(4,3,"CAMERA");
  
  /* P = [y(1:4)';y(5:8)';z(1:3)' y(9)]; */
  MREF(P,0,0) = MREF(y,0,0);
  MREF(P,0,1) = MREF(y,1,0);
  MREF(P,0,2) = MREF(y,2,0);
  MREF(P,0,3) = MREF(y,3,0);
  MREF(P,1,0) = MREF(y,4,0);
  MREF(P,1,1) = MREF(y,5,0);
  MREF(P,1,2) = MREF(y,6,0);
  MREF(P,1,3) = MREF(y,7,0);
  MREF(P,2,0) = MREF(z,0,0);
  MREF(P,2,1) = MREF(z,1,0);
  MREF(P,2,2) = MREF(z,2,0);
  MREF(P,2,3) = MREF(y,8,0);

#ifdef VERBOSE
  /* disp('The computed camera matrix is ') */
  /* disp(P) */
  printf("the camera matrix is:\n");
  matrix_print(stdout,P);
#endif

  return(P);
}

FLOATTYPE testCalib(P,t1,t2,t3,x,y)
     matrix_type *P;
     FLOATTYPE t1,t2,t3,*x,*y;
{
  /* function tt = testCalib(p,pt) */
  matrix_type *point = make_matrix_type(4,1,"GLOBAL_POINT");
  matrix_type *pixel = make_matrix_type(1,3,"SCREEN_POINT");

  /* hpt = [pt 1]; */
  MREF(point,0,0) = t1;
  MREF(point,0,1) = t2;
  MREF(point,0,2) = t3;
  MREF(point,0,3) = 1.0;

  /* tt = p * hpt; */
  matrix_multiply_transpose(P,point,pixel);
  
  /* tt = tt/tt(3); */
  *x = MREF(pixel,0,0) / MREF(pixel,2,0);
  *y = MREF(pixel,1,0) / MREF(pixel,2,0);
  
#ifdef VERBOSE
  matrix_print(stdout,point);
  matrix_print(stdout,pixel);
#endif

  free_matrix_type(pixel);
  free_matrix_type(point);
}

FLOATTYPE find_maxerr(P,test)
     matrix_type *P;
     matrix_type *test;
{
  /* % sanity check to make sure matrix does what it claims */
  
  int i;
  matrix_type *mag = make_matrix_type(1,matrix_height(test),"MAG");
  
  /* for i=1:length(test) */
  
  for(i=0;i<matrix_height(test);i++)
    {
      /*   err = testCalib(P,test(i,1:3)); */
      FLOATTYPE x,y;
      FLOATTYPE err = testCalib(P,
				MREF(test,i,0),
				MREF(test,i,1),
				MREF(test,i,2),
				&x,&y);
      
      /*   er2 = err(1:2)' - test(i,4:5); */
      FLOATTYPE dx = x - MREF(test,i,3);
      FLOATTYPE dy = y - MREF(test,i,4);
      
      /*   mag(i) = sum(er2 .* er2); */
      MREF(mag,i,0) = dx*dx + dy*dy;
      /* end */
    }
  
  /* maxerr = max(mag(i)); */
  {
    FLOATTYPE max = 0.0;
    int i;

    for(i=0;i<matrix_height(mag);i++) 
      {
	if (max < MREF(mag,0,i)) max = MREF(mag,0,i);
      }
    
#ifdef VERBOSE
    printf("maxerr is: %g\n",max);
#endif
    return (max);
  }
}

matrix_type *make_camera(P)
     matrix_type *P;
{     
  /* % P is the camera matrix */

  matrix_type *camera = make_matrix_type(4,4,"CAMERA");
  
  /* P = [y(1:4)';y(5:8)';z(1:3)' y(9)]; */
  MREF(camera,0,0) = MREF(P,0,0);
  MREF(camera,0,1) = MREF(P,0,1);
  MREF(camera,0,2) = MREF(P,0,2);
  MREF(camera,0,3) = MREF(P,0,3);
  MREF(camera,1,0) = MREF(P,1,0);
  MREF(camera,1,1) = MREF(P,1,1);
  MREF(camera,1,2) = MREF(P,1,2);
  MREF(camera,1,3) = MREF(P,1,3);
  MREF(camera,2,0) = MREF(P,2,0);
  MREF(camera,2,1) = MREF(P,2,1);
  MREF(camera,2,2) = MREF(P,2,2);
  MREF(camera,2,3) = MREF(P,2,3);
  MREF(camera,3,0) = (FLOATTYPE) 0.0;
  MREF(camera,3,1) = (FLOATTYPE) 0.0;
  MREF(camera,3,2) = (FLOATTYPE) 0.0;
  MREF(camera,3,3) = (FLOATTYPE) 1.0;

  return(camera);
}

float do_calibration(n,test,camera)
     int n;
     matrix_type *test;
     matrix_type **camera;     
{
  /* a = buildMatrix(test); */
  matrix_type *a = buildMatrix(n,test);
  
  /* c = extractC(a); */
  matrix_type *c = extractC(a);
  
  /* d = extractD(a); */
  matrix_type *d = extractD(a);
  
  /* ctci = (c' * c)^(-1); */
  matrix_type *ctci= make_ctci(c);
  
  /* ea = (d'*d - d'*c*(ctci*c'*d) */
  matrix_type *eigsource= make_eigsource(c,d,ctci);
  
  /* % find eigen-values to arrive at z */
  matrix_type *z = find_z(eigsource);
  
  /* % now find y */
  /* y = - ctci*c'*d*z; */
  matrix_type *y = find_y(ctci,c,d,z);    
  
  /* % P is the camera matrix */
  matrix_type *P = find_camera(y,z);
  
  /* % sanity check to make sure matrix does what it claims */
  FLOATTYPE maxerr = find_maxerr(P,test);
  
  *camera = make_camera(P);

  free_matrix_type(a);
  free_matrix_type(c);
  free_matrix_type(d);
  free_matrix_type(ctci);
  free_matrix_type(eigsource);
  free_matrix_type(z);
  free_matrix_type(y);
  free_matrix_type(P);

  return ((float) maxerr);
}

