/*
 * Copyright (c) 1991-94 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the name of Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
 * POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*--------------------------------------------------------------
 *  This is an example from the Inventor Mentor,
 *  chapter 7, example 1.
 *
 *  This example displays a textured cube (default 
 *  texture coords).
 *------------------------------------------------------------*/

#include <Inventor/Xt/SoXt.h>
#include <Inventor/Sb.h>
#include <Inventor/SoInput.h>
#include <Inventor/SoDB.h>
#include <Inventor/Xt/viewers/SoXtExaminerViewer.h>
#include <Inventor/nodes/SoCube.h>
#include <Inventor/nodes/SoSphere.h>
#include <Inventor/nodes/SoBaseColor.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoTexture2.h>
#include <Inventor/nodes/SoTextureCoordinate2.h>
#include <Inventor/nodes/SoTextureCoordinateBinding.h>
#include <Inventor/nodes/SoCallback.h>
#include <Inventor/nodes/SoOrthographicCamera.h>
#include <Inventor/nodes/SoPerspectiveCamera.h>
#include <Inventor/events/SoMouseButtonEvent.h>
#include <Inventor/nodes/SoEventCallback.h>
#include <Inventor/events/SoKeyboardEvent.h>
#include <Inventor/events/SoEvents.h>
#include <Inventor/projectors/SbPlaneProjector.h>
#include <iostream.h>
#include <Inventor/sensors/SoSensor.h>
#include <Inventor/sensors/SoIdleSensor.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/sensors/SoSensor.h>
#include <Inventor/sensors/SoIdleSensor.h>
#include <Inventor/sensors/SoTimerSensor.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoNormal.h>
#include <Inventor/nodes/SoNormalBinding.h>
#include <Inventor/nodes/SoFaceSet.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <stdio.h>
#include "AliveCamera.h"

SoXtRenderArea *myRenderArea=NULL;
SoCamera *myCamera=NULL;
SoSeparator *root=NULL;
SoTimerSensor *sensor = NULL;
SoCallback *grabNode = NULL;

AliveCamera *cameraMapping;
Widget myWindow;


static void myCallback( void *data, SoSensor *sensor )
{
    static int t = 0;
    int i;
    myRenderArea->render();
    //sensor->schedule();
}
void keyboardCallback(void *ud, SoEventCallback *eventCB)
{
    const SoEvent *ev = eventCB->getEvent();
}

void
cleanup(...)
{
  exit(0);
}

SoCamera *buildCamera(SoXtRenderArea *renderArea)
{
    myCamera = new SoPerspectiveCamera;
    cameraMapping = new AliveCamera(myCamera, renderArea);
    cameraMapping->readCameraParametersFromFile("cameraParameters");
    return myCamera;
}

SoSeparator *buildWorldModelFromFile()
{
    SoInput mySceneInput;
    SoNode *fn;
    if (!mySceneInput.openFile("calib.iv")) {
       printf("Cannot open file\n");
       return NULL;
    }
    SoDB::read(&mySceneInput, fn);
    return (SoSeparator *)fn;
}

SoSeparator *buildWorldModel()
{
  SoSeparator *world = new SoSeparator;
  SoBaseColor *bc = new SoBaseColor;
  SoTransform *xf;
  SoCube *cube = new SoCube;
  SoSeparator *objsep;
  SoSphere *sphere;
    
  bc->rgb.setValue(0, 0, .7);
  world->addChild(bc);

  // floor 
  objsep = new SoSeparator;
  xf = new SoTransform;
  bc = new SoBaseColor;
  // bc->rgb.setValue(.4, .4, .4);
  objsep->addChild(bc);
  cube = new SoCube;
  cube->depth = 12*12;
  cube->height = .09;
  cube->width = 9*12;
  // correct?
    xf->translation.setValue(SbVec3f(0.0, .001, 38));
  objsep->addChild(xf);
  objsep->addChild(cube);
  world->addChild(objsep);
  
  // carpet
  //  objsep = new SoSeparator;
  // xf = new SoTransform;
  // cube = new SoCube;
  // cube->depth = 108;
  // cube->height = .1;
  // cube->width = 120;
  // objsep->addChild(xf);
  // objsep->addChild(cube);
  // world->addChild(objsep);
  
  // back wall 
  objsep = new SoSeparator;
  xf = new SoTransform;
  bc = new SoBaseColor;
  bc->rgb.setValue(0, .6, 0);
  objsep->addChild(bc);
  xf->translation.setValue(0, 60, -58);
  cube = new SoCube;
  cube->width = 120;
  cube->height = 120;
  cube->depth = .1;
  objsep->addChild(xf);
  objsep->addChild(cube);
  world->addChild(objsep);
  
  
  // back patch on door 
  objsep = new SoSeparator;
  xf = new SoTransform;
  bc = new SoBaseColor;
  bc->rgb.setValue(.2, .2, .2);
  objsep->addChild(bc);
  xf->translation.setValue(-60, 60, -62);
  cube = new SoCube;
  cube->width = 1;
  cube->height = 1;
  cube->depth = .1;
  objsep->addChild(xf);
  objsep->addChild(cube);
  world->addChild(objsep);
  
  SoSeparator *boxsep = new SoSeparator;
  
  // box 1 
  objsep = new SoSeparator;
  xf = new SoTransform;
  xf->translation.setValue(-54 + 12, 15, -34+23/2);
  cube = new SoCube;
  cube->width = 24;
  cube->height = 30;
  cube->depth = 23;
  objsep->addChild(xf);
  objsep->addChild(cube);
  boxsep->addChild(objsep);
  
  // box 2 
  objsep = new SoSeparator;
  xf = new SoTransform;
  xf->translation.setValue(54 - 12, 25.5/2, -34+6);
  cube = new SoCube;
  cube->width = 24;
  cube->height = 25.5;
  cube->depth = 12;
  objsep->addChild(xf);
  objsep->addChild(cube);
  boxsep->addChild(objsep);
  
  // box 3 
  objsep = new SoSeparator;
  xf = new SoTransform;
  xf->translation.setValue(3, 16, 38-7);
  cube = new SoCube;
  cube->width = 6;
  cube->height = 32;
  cube->depth = 14;
  objsep->addChild(xf);
  objsep->addChild(cube);
  boxsep->addChild(objsep);
  world->addChild(boxsep);
  
  // backwall mark
  objsep = new SoSeparator;
  xf = new SoTransform;
  xf->translation.setValue(54, 48, -58);
  sphere = new SoSphere; 
  sphere->radius = 2;
  sphere->radius = 3;
  objsep->addChild(xf);
  objsep->addChild(sphere);
  boxsep->addChild(objsep);
  world->addChild(boxsep);
  
  return world;
}

void

main(int , char **argv)
{

   static int attributeList[] = {
	GLX_RGBA,
	GLX_RED_SIZE, 1,
	GLX_GREEN_SIZE, 1,
	GLX_BLUE_SIZE, 1,
	GLX_DEPTH_SIZE, 1,
	GLX_DOUBLEBUFFER,
	GLX_ACCUM_RED_SIZE,  4,
	GLX_ACCUM_GREEN_SIZE,  4,
	GLX_ACCUM_BLUE_SIZE,  4,
	None,
    };

   XVisualInfo *vis = NULL;
    
   myWindow = SoXt::init(argv[0]);
   if(myWindow == NULL) exit(1);

   SoXtExaminerViewer *myViewer = new SoXtExaminerViewer(myWindow);
 
   vis = AliveCamera::buildEnabledVisual(attributeList);
   if(vis)
	myViewer->setNormalVisual(vis); 
   
   root = new SoSeparator;
   root->ref();
   myViewer->setAutoClipping(0);
   myViewer->setAutoRedraw(0);
   root->addChild(buildCamera(myViewer));
   root->addChild(buildWorldModelFromFile());
   SoEventCallback *cbn = new SoEventCallback;
   cbn->addEventCallback(SoKeyboardEvent::getClassTypeId(),
	keyboardCallback, NULL);
   root->addChild(cbn); 
   
   SoSeparator *markers = new SoSeparator;
   root->addChild(markers);
   cameraMapping->setShowMouseMarkers(1, markers);
   cameraMapping->trackMouseEvents();
  
   myViewer->setSceneGraph(root);
   myViewer->setTitle("Simple App");
   myViewer->setDecoration(0);
   myViewer->show();
 
   myRenderArea = myViewer;
   sensor = new SoTimerSensor( myCallback, NULL);
   sensor->setInterval(.02);
   
   // add contour server
   signal(SIGINT,cleanup);
   
   sensor->schedule();
   
   SoXt::show(myWindow);
  

   SoXt::mainLoop();
}
